@@ -1,9 +1,8 @@
-import appdaemon.utils as utils
 from appdaemon.appdaemon import AppDaemon
 import functools
 
-class Futures:
 
+class Futures:
     def __init__(self, ad: AppDaemon):
         self.AD = ad
 
@@ -15,17 +14,17 @@ class Futures:
             self.futures[name] = []
 
         self.futures[name].append(f)
-        self.AD.logger.debug('Registered a future in {}: {}'.format(name, f))
+        self.AD.logger.debug("Registered a future in {}: {}".format(name, f))
 
     def remove_future(self, name, f):
         if name in self.futures:
             self.futures[name].remove(f)
-        
-        self.AD.logger.debug('Future removed from registry {}'.format(f))
+
+        self.AD.logger.debug("Future removed from registry {}".format(f))
 
         if f.cancelled():
             return
-        
+
         if not f.done():
             f.cancel()
 
@@ -35,5 +34,5 @@ class Futures:
 
         for f in self.futures[name]:
             if not f.done() and not f.cancelled():
-                self.AD.logger.debug('Cancelling Future {}'.format(f))
+                self.AD.logger.debug("Cancelling Future {}".format(f))
                 f.cancel()
