@@ -15,7 +15,7 @@ from jinja2 import Environment, FileSystemLoader, select_autoescape
 
 import appdaemon.dashboard as addashboard
 import appdaemon.utils as utils
-import appdaemon.stream as stream
+import appdaemon.stream.adstream as stream
 import appdaemon.admin as adadmin
 
 from appdaemon.appdaemon import AppDaemon
@@ -33,7 +33,9 @@ def securedata(myfunc):
         if self.password is None:
             return await myfunc(*args)
         elif "adcreds" in request.cookies:
-            match = await utils.run_in_executor(self, bcrypt.checkpw, str.encode(self.password), str.encode(request.cookies["adcreds"]))
+            match = await utils.run_in_executor(
+                self, bcrypt.checkpw, str.encode(self.password), str.encode(request.cookies["adcreds"]),
+            )
             if match:
                 return await myfunc(*args)
         elif ("x-ad-access" in request.headers) and (request.headers["x-ad-access"] == self.password):
@@ -59,9 +61,9 @@ def secure(myfunc):
             return await myfunc(*args)
         else:
             if "adcreds" in request.cookies:
-                match = await utils.run_in_executor(self, bcrypt.checkpw,
-                                                    str.encode(self.password),
-                                                    str.encode(request.cookies["adcreds"]))
+                match = await utils.run_in_executor(
+                    self, bcrypt.checkpw, str.encode(self.password), str.encode(request.cookies["adcreds"]),
+                )
                 if match:
                     return await myfunc(*args)
                 else:
@@ -72,8 +74,35 @@ def secure(myfunc):
     return wrapper
 
 
-class HTTP:
+def route_secure(myfunc):
+    """
+    Take care of streams and service calls
+    """
 
+    async def wrapper(*args):
+
+        self = args[0]
+        request = args[1]
+        if self.password is None or self.valid_tokens == []:
+            return await myfunc(*args)
+
+        elif "adcreds" in request.cookies:
+            match = await utils.run_in_executor(
+                self, bcrypt.checkpw, str.encode(self.password), str.encode(request.cookies["adcreds"])
+            )
+            if match:
+                return await myfunc(*args)
+
+        elif "token" in request.query and request.query["token"] in self.valid_tokens:
+            return await myfunc(*args)
+
+        else:
+            return self.get_response(request, "401", "Unauthorized")
+
+    return wrapper
+
+
+class HTTP:
     def __init__(self, ad: AppDaemon, loop, logging, appdaemon, dashboard, admin, api, http):
 
         self.AD = ad
@@ -92,30 +121,24 @@ class HTTP:
         self.template_dir = os.path.join(os.path.dirname(__file__), "assets", "templates")
 
         self.password = None
-        self._process_arg("password", http)
-
+        self.valid_tokens = []
         self.url = None
-        self._process_arg("url", http)
-
-        self.work_factor = 8
-        self._process_arg("work_factor", http)
-
+        self.work_factor = 12
         self.ssl_certificate = None
-        self._process_arg("ssl_certificate", http)
-
         self.ssl_key = None
-        self._process_arg("ssl_key", http)
-
         self.transport = "ws"
-        self._process_arg("transport", http)
-        self.logger.info("Using '%s' for event stream", self.transport)
 
         self.config_dir = None
         self._process_arg("config_dir", dashboard)
 
+        self.static_dirs = {}
+
+        self._process_http(http)
+
         self.stopping = False
 
         self.endpoints = {}
+        self.app_routes = {}
 
         self.dashboard_obj = None
         self.admin_obj = None
@@ -154,12 +177,11 @@ class HTTP:
             self.loop = loop
             self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)
 
-            #TODO the `context` local varialble is never used after its initialization, maybe it can be removed
             if self.ssl_certificate is not None and self.ssl_key is not None:
-                context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
-                context.load_cert_chain(self.ssl_certificate, self.ssl_key)
+                self.context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
+                self.context.load_cert_chain(self.ssl_certificate, self.ssl_key)
             else:
-                context = None
+                self.context = None
 
             self.setup_http_routes()
 
@@ -183,15 +205,19 @@ class HTTP:
                 self.stats_update = "realtime"
                 self._process_arg("stats_update", admin)
 
-                self.admin_obj = adadmin.Admin(self.config_dir, logging, self.AD,
-                                               javascript_dir=self.javascript_dir,
-                                               template_dir=self.template_dir,
-                                               css_dir=self.css_dir,
-                                               fonts_dir=self.fonts_dir,
-                                               webfonts_dir=self.webfonts_dir,
-                                               images_dir=self.images_dir,
-                                               **admin
-                                               )
+                self.admin_obj = adadmin.Admin(
+                    self.config_dir,
+                    logging,
+                    self.AD,
+                    javascript_dir=self.javascript_dir,
+                    template_dir=self.template_dir,
+                    css_dir=self.css_dir,
+                    fonts_dir=self.fonts_dir,
+                    webfonts_dir=self.webfonts_dir,
+                    images_dir=self.images_dir,
+                    transport=self.transport,
+                    **admin
+                )
 
             else:
                 self.logger.info("Admin Interface is disabled")
@@ -201,7 +227,7 @@ class HTTP:
 
             if dashboard is not None:
                 self.logger.info("Starting Dashboards")
-                
+
                 self._process_arg("dashboard_dir", dashboard)
 
                 self.compile_on_start = True
@@ -222,7 +248,7 @@ class HTTP:
                 if "rss_feeds" in dashboard:
                     self.rss_feeds = []
                     for feed in dashboard["rss_feeds"]:
-                        if feed["target"].count('.') != 1:
+                        if feed["target"].count(".") != 1:
                             self.logger.warning("Invalid RSS feed target: %s", feed["target"])
                         else:
                             self.rss_feeds.append(feed)
@@ -255,19 +281,22 @@ class HTTP:
                 else:
                     self.compile_dir = os.path.join(self.config_dir, "compiled")
 
-                self.dashboard_obj = addashboard.Dashboard(self.config_dir, self.logging,
-                                                           dash_compile_on_start=self.compile_on_start,
-                                                           dash_force_compile=self.force_compile,
-                                                           profile_dashboard=self.profile_dashboard,
-                                                           dashboard_dir=self.dashboard_dir,
-                                                           fa4compatibility=self.fa4compatibility,
-                                                           transport=self.transport,
-                                                           javascript_dir=self.javascript_dir,
-                                                           template_dir=self.template_dir,
-                                                           css_dir=self.css_dir,
-                                                           fonts_dir=self.fonts_dir,
-                                                           webfonts_dir=self.webfonts_dir,
-                                                           images_dir=self.images_dir)
+                self.dashboard_obj = addashboard.Dashboard(
+                    self.config_dir,
+                    self.logging,
+                    dash_compile_on_start=self.compile_on_start,
+                    dash_force_compile=self.force_compile,
+                    profile_dashboard=self.profile_dashboard,
+                    dashboard_dir=self.dashboard_dir,
+                    fa4compatibility=self.fa4compatibility,
+                    transport=self.transport,
+                    javascript_dir=self.javascript_dir,
+                    template_dir=self.template_dir,
+                    css_dir=self.css_dir,
+                    fonts_dir=self.fonts_dir,
+                    webfonts_dir=self.webfonts_dir,
+                    images_dir=self.images_dir,
+                )
                 self.setup_dashboard_routes()
 
             else:
@@ -277,20 +306,39 @@ class HTTP:
             # Finish up and start the server
             #
 
-            #handler = self.app.make_handler()
+            # handler = self.app.make_handler()
 
-            #f = loop.create_server(handler, "0.0.0.0", int(self.port), ssl=context)
-            #loop.create_task(f)
+            # f = loop.create_server(handler, "0.0.0.0", int(self.port), ssl=context)
+            # loop.create_task(f)
 
             if self.dashboard_obj is not None:
                 loop.create_task(self.update_rss())
 
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in HTTP module")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
+
+    def _process_http(self, http):
+        self._process_arg("password", http)
+        self._process_arg("tokens", http)
+        self._process_arg("work_factor", http)
+        self._process_arg("ssl_certificate", http)
+        self._process_arg("ssl_key", http)
+
+        self._process_arg("url", http)
+        if not self.url:
+            self.logger.warning(
+                "'{arg}' is '{value}'. Please configure appdaemon.yaml".format(arg="url", value=self.url)
+            )
+            exit(0)
+
+        self._process_arg("transport", http)
+        self.logger.info("Using '%s' for event stream", self.transport)
+
+        self._process_arg("static_dirs", http)
 
     async def start_server(self):
 
@@ -298,18 +346,18 @@ class HTTP:
 
         self.runner = web.AppRunner(self.app)
         await self.runner.setup()
-        site = web.TCPSite(self.runner, '0.0.0.0', int(self.port))
+        site = web.TCPSite(self.runner, "0.0.0.0", int(self.port), ssl_context=self.context)
         await site.start()
 
     async def stop_server(self):
         self.logger.info("Shutting down webserver")
         #
-        # We sjould do this nut it makes AD hang so ...
+        # We should do this but it makes AD hang so ...
         #
-        #await self.runner.cleanup()
+        # await self.runner.cleanup()
 
     async def add_response_headers(self, request, response):
-        for header, value in self.http['headers'].items():
+        for header, value in self.http["headers"].items():
             response.headers[header] = value
 
     def stop(self):
@@ -334,7 +382,7 @@ class HTTP:
             password = data["password"]
 
             if password == self.password:
-                self.access.info("Succesful logon from %s", request.host)
+                self.access.info("Successful logon from %s", request.host)
                 hashed = bcrypt.hashpw(str.encode(self.password), bcrypt.gensalt(self.work_factor))
                 if self.admin is not None:
                     response = await self._admin_page(request)
@@ -350,12 +398,12 @@ class HTTP:
                 response = await self.logon_page(request)
 
             return response
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in logon_response()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Server error in logon_response()")
 
     # noinspection PyUnusedLocal
@@ -369,11 +417,11 @@ class HTTP:
 
     @secure
     async def load_dash(self, request):
-        name = request.match_info.get('name', "Anonymous")
+        name = request.match_info.get("name", "Anonymous")
         params = request.query
         skin = params.get("skin", "default")
         recompile = params.get("recompile", False)
-        if recompile == '1':
+        if recompile == "1":
             recompile = True
 
         response = await utils.run_in_executor(self, self.dashboard_obj.get_dashboard, name, skin, recompile)
@@ -391,7 +439,9 @@ class HTTP:
                         for feed_data in self.rss_feeds:
                             feed = await utils.run_in_executor(self, feedparser.parse, feed_data["feed"])
                             if "bozo_exception" in feed:
-                                self.logger.warning("Error in RSS feed %s: %s", feed_data["feed"], feed["bozo_exception"])
+                                self.logger.warning(
+                                    "Error in RSS feed %s: %s", feed_data["feed"], feed["bozo_exception"],
+                                )
                             else:
                                 new_state = {"feed": feed}
 
@@ -399,12 +449,12 @@ class HTTP:
                                 await self.AD.state.set_state("rss", "admin", feed_data["target"], state=new_state)
 
                     await asyncio.sleep(1)
-                except:
-                    self.logger.warning('-' * 60)
+                except Exception:
+                    self.logger.warning("-" * 60)
                     self.logger.warning("Unexpected error in update_rss()")
-                    self.logger.warning('-' * 60)
+                    self.logger.warning("-" * 60)
                     self.logger.warning(traceback.format_exc())
-                    self.logger.warning('-' * 60)
+                    self.logger.warning("-" * 60)
 
     #
     # REST API
@@ -419,8 +469,8 @@ class HTTP:
         namespace = None
         entity_id = None
         try:
-            entity_id = request.match_info.get('entity')
-            namespace = request.match_info.get('namespace')
+            entity_id = request.match_info.get("entity")
+            namespace = request.match_info.get("namespace")
 
             self.logger.debug("get_state() called, ns=%s, entity=%s", namespace, entity_id)
             state = self.AD.state.get_entity(namespace, entity_id)
@@ -428,20 +478,20 @@ class HTTP:
             self.logger.debug("result = %s", state)
 
             return web.json_response({"state": state}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_entity()")
             self.logger.warning("Namespace: %s, entity: %s", namespace, entity_id)
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_entity()")
 
     @securedata
     async def get_namespace(self, request):
         namespace = None
         try:
-            namespace = request.match_info.get('namespace')
+            namespace = request.match_info.get("namespace")
 
             self.logger.debug("get_namespace() called, ns=%s", namespace)
             state = self.AD.state.get_entity(namespace)
@@ -452,13 +502,13 @@ class HTTP:
                 return self.get_response(request, 404, "Namespace Not Found")
 
             return web.json_response({"state": state}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_namespace()")
             self.logger.warning("Namespace: %s", namespace)
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_namespace()")
 
     @securedata
@@ -466,7 +516,7 @@ class HTTP:
 
         namespace = None
         try:
-            namespace = request.match_info.get('namespace')
+            namespace = request.match_info.get("namespace")
 
             self.logger.debug("get_namespace_entities() called, ns=%s", namespace)
             state = self.AD.state.list_namespace_entities(namespace)
@@ -477,13 +527,13 @@ class HTTP:
                 return self.get_response(request, 404, "Namespace Not Found")
 
             return web.json_response({"state": state}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_namespace_entities()")
             self.logger.warning("Namespace: %s", namespace)
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_namespace_entities()")
 
     @securedata
@@ -495,15 +545,14 @@ class HTTP:
             self.logger.debug("result = %s", state)
 
             return web.json_response({"state": state}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_namespaces()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_namespaces()")
 
-
     @securedata
     async def get_services(self, request):
 
@@ -513,15 +562,14 @@ class HTTP:
             self.logger.debug("result = %s", state)
 
             return web.json_response({"state": state}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_services()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_services()")
 
-
     @securedata
     async def get_state(self, request):
         try:
@@ -534,12 +582,12 @@ class HTTP:
             self.logger.debug("result = %s", state)
 
             return web.json_response({"state": state}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_state()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_state()")
 
     @securedata
@@ -550,12 +598,12 @@ class HTTP:
             logs = await utils.run_in_executor(self, self.AD.logging.get_admin_logs)
 
             return web.json_response({"logs": logs}, dumps=utils.convert_json)
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in get_logs()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return self.get_response(request, 500, "Unexpected error in get_logs()")
 
     # noinspection PyUnusedLocal
@@ -568,9 +616,9 @@ class HTTP:
                 return self.get_response(request, 400, "JSON Decode Error")
 
             args = {}
-            namespace = request.match_info.get('namespace')
-            domain = request.match_info.get('domain')
-            service = request.match_info.get('service')
+            namespace = request.match_info.get("namespace")
+            domain = request.match_info.get("domain")
+            service = request.match_info.get("service")
             #
             # Some value munging for dashboard
             #
@@ -578,22 +626,22 @@ class HTTP:
                 if key == "service":
                     pass
                 elif key == "rgb_color":
-                    m = re.search('\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)', data[key])
+                    m = re.search(r"\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)", data[key])
                     if m:
                         r = m.group(1)
                         g = m.group(2)
                         b = m.group(3)
                         args["rgb_color"] = [r, g, b]
                 elif key == "xy_color":
-                    m = re.search('\s*(\d+\.\d+)\s*,\s*(\d+\.\d+)', data[key])
+                    m = re.search(r"\s*(\d+\.\d+)\s*,\s*(\d+\.\d+)", data[key])
                     if m:
                         x = m.group(1)
                         y = m.group(2)
                         args["xy_color"] = [x, y]
                 elif key == "json_args":
-                      json_args = json.loads(data[key])
-                      for k in json_args.keys():
-                         args[k] = json_args[k]
+                    json_args = json.loads(data[key])
+                    for k in json_args.keys():
+                        args[k] = json_args[k]
                 else:
                     args[key] = data[key]
 
@@ -602,14 +650,14 @@ class HTTP:
             await self.AD.services.call_service(namespace, domain, service, args)
             return web.Response(status=200)
 
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in call_service()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return web.Response(status=500)
-    
+
     @securedata
     async def fire_event(self, request):
         try:
@@ -619,8 +667,8 @@ class HTTP:
                 return self.get_response(request, 400, "JSON Decode Error")
 
             args = {}
-            namespace = request.match_info.get('namespace')
-            event = request.match_info.get('event')
+            namespace = request.match_info.get("namespace")
+            event = request.match_info.get("event")
             #
             # Some value munging for dashboard
             #
@@ -637,12 +685,12 @@ class HTTP:
 
             return web.Response(status=200)
 
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error in fire_event()")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             return web.Response(status=500)
 
     # noinspection PyUnusedLocal
@@ -652,88 +700,145 @@ class HTTP:
     # Stream Handling
 
     async def stream_update(self, namespace, data):
-        #self.logger.debug("stream_update() %s:%s", namespace, data)
+        # self.logger.debug("stream_update() %s:%s", namespace, data)
         data["namespace"] = namespace
-        self.AD.thread_async.call_async_no_wait(self.stream.send_update, data)
+        self.AD.thread_async.call_async_no_wait(self.stream.process_event, data)
 
     # Routes, Status and Templates
 
     def setup_api_routes(self):
-        self.app.router.add_post('/api/appdaemon/service/{namespace}/{domain}/{service}', self.call_service)
-        self.app.router.add_post('/api/appdaemon/event/{namespace}/{event}', self.fire_event)
-        self.app.router.add_get('/api/appdaemon/service/', self.get_services)
-        self.app.router.add_get('/api/appdaemon/state/{namespace}/{entity}', self.get_entity)
-        self.app.router.add_get('/api/appdaemon/state/{namespace}', self.get_namespace)
-        self.app.router.add_get('/api/appdaemon/state/{namespace}/', self.get_namespace_entities)
-        self.app.router.add_get('/api/appdaemon/state/', self.get_namespaces)
-        self.app.router.add_get('/api/appdaemon/state', self.get_state)
-        self.app.router.add_get('/api/appdaemon/logs', self.get_logs)
-        self.app.router.add_post('/api/appdaemon/{app}', self.call_api)
-        self.app.router.add_get('/api/appdaemon', self.get_ad)
+        self.app.router.add_post("/api/appdaemon/service/{namespace}/{domain}/{service}", self.call_service)
+        self.app.router.add_post("/api/appdaemon/event/{namespace}/{event}", self.fire_event)
+        self.app.router.add_get("/api/appdaemon/service/", self.get_services)
+        self.app.router.add_get("/api/appdaemon/state/{namespace}/{entity}", self.get_entity)
+        self.app.router.add_get("/api/appdaemon/state/{namespace}", self.get_namespace)
+        self.app.router.add_get("/api/appdaemon/state/{namespace}/", self.get_namespace_entities)
+        self.app.router.add_get("/api/appdaemon/state/", self.get_namespaces)
+        self.app.router.add_get("/api/appdaemon/state", self.get_state)
+        self.app.router.add_get("/api/appdaemon/logs", self.get_logs)
+        self.app.router.add_post("/api/appdaemon/{app}", self.call_api)
+        self.app.router.add_get("/api/appdaemon", self.get_ad)
 
     def setup_http_routes(self):
-        self.app.router.add_get('/favicon.ico', self.not_found)
-        self.app.router.add_get('/{gfx}.png', self.not_found)
-        self.app.router.add_post('/logon_response', self.logon_response)
+        self.app.router.add_get("/favicon.ico", self.not_found)
+        self.app.router.add_get("/{gfx}.png", self.not_found)
+        self.app.router.add_post("/logon_response", self.logon_response)
 
         # Add static path for JavaScript
-        self.app.router.add_static('/javascript', self.javascript_dir)
+        self.app.router.add_static("/javascript", self.javascript_dir)
 
         # Add static path for fonts
-        self.app.router.add_static('/fonts', self.fonts_dir)
+        self.app.router.add_static("/fonts", self.fonts_dir)
 
         # Add static path for webfonts
-        self.app.router.add_static('/webfonts', self.webfonts_dir)
+        self.app.router.add_static("/webfonts", self.webfonts_dir)
 
         # Add static path for images
-        self.app.router.add_static('/images', self.images_dir)
+        self.app.router.add_static("/images", self.images_dir)
 
         # Add static path for css
-        self.app.router.add_static('/css', self.css_dir)
+        self.app.router.add_static("/css", self.css_dir)
 
         if self.admin is not None:
-            self.app.router.add_get('/', self.admin_page)
+            self.app.router.add_get("/", self.admin_page)
         elif self.dashboard is not None:
-            self.app.router.add_get('/', self.list_dash)
+            self.app.router.add_get("/", self.list_dash)
         else:
-            self.app.router.add_get('/', self.error_page)
+            self.app.router.add_get("/", self.error_page)
+
+        #
+        # For App based Web Server
+        #
+        self.app.router.add_get("/app/{route}", self.app_webserver)
+
+        #
+        # Add static path for apps
+        #
+        apps_static = os.path.join(self.AD.config_dir, "www")
+        exists = True
+
+        if not os.path.isdir(apps_static):  # check if the folder exists
+            try:
+                os.mkdir(apps_static)
+            except OSError:
+                self.logger.warning("Creation of the Web directory %s failed", apps_static)
+                exists = False
+            else:
+                self.logger.debug("Successfully created the Web directory %s ", apps_static)
+
+        if exists:
+            self.app.router.add_static("/local", apps_static)
+        #
+        # Setup user defined static paths
+        #
+
+        for name, static_dir in self.static_dirs.items():
+            if not os.path.isdir(static_dir):  # check if the folder exists
+                self.logger.warning("The Web directory %s doesn't exist. So static route not set up", static_dir)
+
+            else:
+                self.app.router.add_static("/{}".format(name), static_dir)
+                self.logger.debug("Successfully created the Web directory %s ", static_dir)
 
     def setup_dashboard_routes(self):
-        self.app.router.add_get('/list', self.list_dash)
-        self.app.router.add_get('/{name}', self.load_dash)
+        self.app.router.add_get("/list", self.list_dash)
+        self.app.router.add_get("/{name}", self.load_dash)
 
         # Setup Templates
 
-        self.app.router.add_static('/compiled_javascript', self.dashboard_obj.compiled_javascript_dir)
+        self.app.router.add_static("/compiled_javascript", self.dashboard_obj.compiled_javascript_dir)
 
-        self.app.router.add_static('/compiled_css', self.dashboard_obj.compiled_css_dir)
+        self.app.router.add_static("/compiled_css", self.dashboard_obj.compiled_css_dir)
 
         # Add path for custom_css if it exists
 
         custom_css = os.path.join(self.dashboard_obj.config_dir, "custom_css")
         if os.path.isdir(custom_css):
-            self.app.router.add_static('/custom_css', custom_css)
+            self.app.router.add_static("/custom_css", custom_css)
+
+        # Add path for custom_javascript if it exists
+
+        custom_javascript = os.path.join(self.dashboard_obj.config_dir, "custom_javascript")
+        if os.path.isdir(custom_javascript):
+            self.app.router.add_static("/custom_javascript", custom_javascript)
+
     # API
 
     async def terminate_app(self, name):
         if name in self.endpoints:
             del self.endpoints[name]
 
+        if name in self.app_routes:
+            del self.app_routes[name]
+
     def get_response(self, request, code, error):
-        res = "<html><head><title>{} {}</title></head><body><h1>{} {}</h1>Error in API Call</body></html>".format(code, error, code, error)
-        app = request.match_info.get('app', "system")
+        res = "<html><head><title>{} {}</title></head><body><h1>{} {}</h1>Error in API Call</body></html>".format(
+            code, error, code, error
+        )
+        app = request.match_info.get("app", "system")
         if code == 200:
             self.access.info("API Call to %s: status: %s", app, code)
         else:
-            self.logger.warning("API Call to %s: status: %s, %s", app, code, error)
+            self.access.warning("API Call to %s: status: %s, %s", app, code, error)
         return web.Response(body=res, status=code)
 
+    def get_web_response(self, request, code, error):
+        res = "<html><head><title>{} {}</title></head><body><h1>{} {}</h1>Error in Web Service Call</body></html>".format(
+            code, error, code, error
+        )
+        app = request.match_info.get("app", "system")
+        if code == 200:
+            self.access.info("Web Call to %s: status: %s", app, code)
+        else:
+            self.access.warning("Web Call to %s: status: %s, %s", app, code, error)
+        return web.Response(text=res, content_type="text/html")
+
     @securedata
     async def call_api(self, request):
 
         code = 200
         ret = ""
-        app = request.match_info.get('app')
+        app = request.match_info.get("app")
 
         try:
             args = await request.json()
@@ -742,12 +847,12 @@ class HTTP:
 
         try:
             ret, code = await self.dispatch_app_by_name(app, args)
-        except:
-            self.logger.warning('-' * 60)
-            self.logger.warning("Unexpected error during API call")
-            self.logger.warning('-' * 60)
-            self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.error("-" * 60)
+            self.logger.error("Unexpected error during API call")
+            self.logger.error("-" * 60)
+            self.logger.error(traceback.format_exc())
+            self.logger.error("-" * 60)
 
         if code == 404:
             return self.get_response(request, 404, "App Not Found")
@@ -785,7 +890,82 @@ class HTTP:
             else:
                 return await utils.run_in_executor(self, callback, args)
         else:
-            return '', 404
+            return "", 404
+
+    #
+    # App based Web Server
+    #
+    async def register_route(self, cb, route, name, **kwargs):
+
+        if not asyncio.iscoroutinefunction(cb):  # must be async function
+            self.logger.warning(
+                "Could not Register Callback for %s, using Route %s as Web Server Route. Callback must be Async",
+                name,
+                route,
+            )
+            return
+
+        handle = uuid.uuid4().hex
+
+        if name not in self.app_routes:
+            self.app_routes[name] = {}
+
+        token = kwargs.get("token")
+        self.app_routes[name][handle] = {"callback": cb, "route": route, "token": token}
+
+        return handle
+
+    async def unregister_route(self, handle, name):
+        if name in self.app_routes and handle in self.app_routes[name]:
+            del self.app_routes[name][handle]
+
+    @route_secure
+    async def app_webserver(self, request):
+
+        name = None
+        route = request.match_info.get("route")
+        token = request.query.get("token")
+
+        code = 404
+        error = "Requested Server does not exist"
+
+        callback = None
+        for name in self.app_routes:
+            if callback is not None:  # a callback has been collected
+                break
+
+            for handle in self.app_routes[name]:
+                app_route = self.app_routes[name][handle]["route"]
+                app_token = self.app_routes[name][handle]["token"]
+
+                if app_route == route:
+                    if app_token is not None and app_token != token:
+                        return self.get_web_response(request, "401", "Unauthorized")
+
+                    callback = self.app_routes[name][handle]["callback"]
+                    break
+
+        if callback is not None:
+            self.access.debug("Web Call to %s for %s", route, name)
+
+            try:
+                f = asyncio.ensure_future(callback(request))
+                self.AD.futures.add_future(name, f)
+                return await f
+            except asyncio.CancelledError:
+                code = 503
+                error = "Request was Cancelled"
+
+            except Exception:
+                self.logger.error("-" * 60)
+                self.logger.error("Unexpected error during Web call")
+                self.logger.error("-" * 60)
+                self.logger.error(traceback.format_exc())
+                self.logger.error("-" * 60)
+                code = 503
+                error = "Request had an Error"
+
+        return self.get_web_response(request, str(code), error)
 
     #
     # Admin
@@ -814,8 +994,7 @@ class HTTP:
             params = {}
 
             env = Environment(
-                loader=FileSystemLoader(self.template_dir),
-                autoescape=select_autoescape(['html', 'xml'])
+                loader=FileSystemLoader(self.template_dir), autoescape=select_autoescape(["html", "xml"]),
             )
 
             template = env.get_template("logon.jinja2")
@@ -823,21 +1002,19 @@ class HTTP:
 
             return rendered_template
 
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error creating logon page")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
 
     def generate_error_page(self, scheme, url):
         try:
             params = {}
 
-
             env = Environment(
-                loader=FileSystemLoader(self.template_dir),
-                autoescape=select_autoescape(['html', 'xml'])
+                loader=FileSystemLoader(self.template_dir), autoescape=select_autoescape(["html", "xml"]),
             )
 
             template = env.get_template("error.jinja2")
@@ -845,10 +1022,9 @@ class HTTP:
 
             return rendered_template
 
-        except:
-            self.logger.warning('-' * 60)
+        except Exception:
+            self.logger.warning("-" * 60)
             self.logger.warning("Unexpected error creating logon page")
-            self.logger.warning('-' * 60)
+            self.logger.warning("-" * 60)
             self.logger.warning(traceback.format_exc())
-            self.logger.warning('-' * 60)
-
+            self.logger.warning("-" * 60)
