@@ -2,7 +2,6 @@ import traceback
 import datetime
 from datetime import timedelta
 import pytz
-import astral
 import random
 import uuid
 import re
@@ -12,10 +11,10 @@ from collections import OrderedDict
 
 import appdaemon.utils as utils
 from appdaemon.appdaemon import AppDaemon
+from astral.location import Location, LocationInfo
 
 
 class Scheduler:
-
     def __init__(self, ad: AppDaemon):
         self.AD = ad
 
@@ -43,7 +42,16 @@ class Scheduler:
         self.set_start_time()
 
         if self.AD.endtime is not None:
-            unaware_end = datetime.datetime.strptime(self.AD.endtime, "%Y-%m-%d %H:%M:%S")
+            unaware_end = None
+            try:
+                unaware_end = datetime.datetime.strptime(self.AD.endtime, "%Y-%m-%d %H:%M:%S")
+            except ValueError:
+                try:
+                    unaware_end = datetime.datetime.strptime(self.AD.endtime, "%Y-%m-%d#%H:%M:%S")
+                except ValueError:
+                    pass
+            if unaware_end is None:
+                raise ValueError("Invalid end time for time travel")
             aware_end = self.AD.tz.localize(unaware_end)
             self.endtime = aware_end.astimezone(pytz.utc)
         else:
@@ -55,9 +63,20 @@ class Scheduler:
 
     def set_start_time(self):
         tt = False
+        unaware_now = None
         if self.AD.starttime is not None:
             tt = True
-            unaware_now = datetime.datetime.strptime(self.AD.starttime, "%Y-%m-%d %H:%M:%S")
+            try:
+                unaware_now = datetime.datetime.strptime(self.AD.starttime, "%Y-%m-%d %H:%M:%S")
+            except ValueError:
+                # Support "#" as date and time separator as well
+                try:
+                    unaware_now = datetime.datetime.strptime(self.AD.starttime, "%Y-%m-%d#%H:%M:%S")
+                except ValueError:
+                    # Catching this allows us to raise a single exception and avoid a nested exception
+                    pass
+            if unaware_now is None:
+                raise ValueError("Invalid start time for time travel")
             aware_now = self.AD.tz.localize(unaware_now)
             self.now = aware_now.astimezone(pytz.utc)
         else:
@@ -73,46 +92,71 @@ class Scheduler:
         self.stopping = True
 
     async def cancel_timer(self, name, handle):
+        executed = False
         self.logger.debug("Canceling timer for %s", name)
-        if name in self.schedule and handle in self.schedule[name]:
+        if self.timer_running(name, handle):
             del self.schedule[name][handle]
             await self.AD.state.remove_entity("admin", "scheduler_callback.{}".format(handle))
+            executed = True
+
         if name in self.schedule and self.schedule[name] == {}:
             del self.schedule[name]
 
+        if not executed:
+            self.logger.warning("Invalid callback handle '{}' in cancel_timer() from app {}".format(handle, name))
+
+        return executed
+
+    def timer_running(self, name, handle):
+        """Check if the handler is valid
+        by ensuring the timer is still running"""
+
+        if name in self.schedule and handle in self.schedule[name]:
+            return True
+
+        return False
+
     # noinspection PyBroadException
     async def exec_schedule(self, name, args, uuid_):
         try:
-            if "inactive" in args:
-                return
             # Call function
             if "__entity" in args["kwargs"]:
                 #
                 # it's a "duration" entry
                 #
-                executed = await self.AD.threading.dispatch_worker(name, {
-                    "id": uuid_,
-                    "name": name,
-                    "objectid": self.AD.app_management.objects[name]["id"],
-                    "type": "state",
-                    "function": args["callback"],
-                    "attribute": args["kwargs"]["__attribute"],
-                    "entity": args["kwargs"]["__entity"],
-                    "new_state": args["kwargs"]["__new_state"],
-                    "old_state": args["kwargs"]["__old_state"],
-                    "pin_app": args["pin_app"],
-                    "pin_thread": args["pin_thread"],
-                    "kwargs": args["kwargs"],
-                })
+
+                # first remove the duration parameter
+                if args["kwargs"].get("__duration"):
+                    del args["kwargs"]["__duration"]
+
+                executed = await self.AD.threading.dispatch_worker(
+                    name,
+                    {
+                        "id": uuid_,
+                        "name": name,
+                        "objectid": self.AD.app_management.objects[name]["id"],
+                        "type": "state",
+                        "function": args["callback"],
+                        "attribute": args["kwargs"]["__attribute"],
+                        "entity": args["kwargs"]["__entity"],
+                        "new_state": args["kwargs"]["__new_state"],
+                        "old_state": args["kwargs"]["__old_state"],
+                        "pin_app": args["pin_app"],
+                        "pin_thread": args["pin_thread"],
+                        "kwargs": args["kwargs"],
+                    },
+                )
 
                 if executed is True:
                     remove = args["kwargs"].get("oneshot", False)
                     if remove is True:
                         await self.AD.state.cancel_state_callback(args["kwargs"]["__handle"], name)
-                        
-                        if "__timeout" in args["kwargs"]: #meaning there is a timeout for this callback
-                            await self.cancel_timer(name, args["kwargs"]["__timeout"]) #cancel it as no more needed
-                            
+
+                        if "__timeout" in args["kwargs"] and self.timer_running(
+                            name, args["kwargs"]["__timeout"]
+                        ):  # meaning there is a timeout for this callback
+                            await self.cancel_timer(name, args["kwargs"]["__timeout"])  # cancel it as no more needed
+
             elif "__state_handle" in args["kwargs"]:
                 #
                 # It's a state timeout entry - just delete the callback
@@ -123,20 +167,28 @@ class Scheduler:
                 # It's an event timeout entry - just delete the callback
                 #
                 await self.AD.events.cancel_event_callback(name, args["kwargs"]["__event_handle"])
+            elif "__log_handle" in args["kwargs"]:
+                #
+                # It's a log timeout entry - just delete the callback
+                #
+                await self.AD.logging.cancel_log_callback(name, args["kwargs"]["__log_handle"])
             else:
                 #
                 # A regular callback
                 #
-                await self.AD.threading.dispatch_worker(name, {
-                    "id": uuid_,
-                    "name": name,
-                    "objectid": self.AD.app_management.objects[name]["id"],
-                    "type": "scheduler",
-                    "function": args["callback"],
-                    "pin_app": args["pin_app"],
-                    "pin_thread": args["pin_thread"],
-                    "kwargs": args["kwargs"]
-                })
+                await self.AD.threading.dispatch_worker(
+                    name,
+                    {
+                        "id": uuid_,
+                        "name": name,
+                        "objectid": self.AD.app_management.objects[name]["id"],
+                        "type": "scheduler",
+                        "function": args["callback"],
+                        "pin_app": args["pin_app"],
+                        "pin_thread": args["pin_thread"],
+                        "kwargs": args["kwargs"],
+                    },
+                )
             # If it is a repeating entry, rewrite with new timestamp
             if args["repeat"]:
                 if args["type"] == "next_rising" or args["type"] == "next_setting":
@@ -150,25 +202,30 @@ class Scheduler:
                     args["timestamp"] = args["basetime"] + timedelta(seconds=self.get_offset(args))
                 # Update entity
 
-                await self.AD.state.set_state("_scheduler", "admin", "scheduler_callback.{}".format(uuid_), execution_time=utils.dt_to_str(args["timestamp"].replace(microsecond=0), self.AD.tz))
+                await self.AD.state.set_state(
+                    "_scheduler",
+                    "admin",
+                    "scheduler_callback.{}".format(uuid_),
+                    execution_time=utils.dt_to_str(args["timestamp"].replace(microsecond=0), self.AD.tz),
+                )
             else:
                 # Otherwise just delete
                 await self.AD.state.remove_entity("admin", "scheduler_callback.{}".format(uuid_))
 
                 del self.schedule[name][uuid_]
 
-        except:
+        except Exception:
             error_logger = logging.getLogger("Error.{}".format(name))
-            error_logger.warning('-' * 60)
+            error_logger.warning("-" * 60)
             error_logger.warning("Unexpected error during exec_schedule() for App: %s", name)
             error_logger.warning("Args: %s", args)
-            error_logger.warning('-' * 60)
+            error_logger.warning("-" * 60)
             error_logger.warning(traceback.format_exc())
-            error_logger.warning('-' * 60)
+            error_logger.warning("-" * 60)
             if self.AD.logging.separate_error_log() is True:
                 self.logger.warning("Logged an error to %s", self.AD.logging.get_filename("error_log"))
             error_logger.warning("Scheduler entry has been deleted")
-            error_logger.warning('-' * 60)
+            error_logger.warning("-" * 60)
             await self.AD.state.remove_entity("admin", "scheduler_callback.{}".format(uuid_))
             del self.schedule[name][uuid_]
 
@@ -182,11 +239,7 @@ class Scheduler:
         if longitude < -180 or longitude > 180:
             raise ValueError("Longitude needs to be -180 .. 180")
 
-        elevation = self.AD.elevation
-
-        self.location = astral.Location((
-            '', '', latitude, longitude, self.AD.tz.zone, elevation
-        ))
+        self.location = Location(LocationInfo("", "", self.AD.tz.zone, latitude, longitude))
 
     def sun(self, type, offset):
         if offset < 0:
@@ -206,27 +259,25 @@ class Scheduler:
         mod = offset
         while True:
             try:
-                next_rising_dt = self.location.sunrise(
-                    (self.now + datetime.timedelta(seconds=offset) + datetime.timedelta(days=mod)).date(), local=False
-                )
+                dt = (self.now + datetime.timedelta(seconds=offset) + datetime.timedelta(days=mod)).date()
+                next_rising_dt = self.location.sunrise(date=dt, local=False, observer_elevation=self.AD.elevation)
                 if next_rising_dt > self.now:
                     break
-            except astral.AstralError:
+            except ValueError:
                 pass
             mod += 1
 
         return next_rising_dt
 
-    def next_sunset(self, offset = 0):
+    def next_sunset(self, offset=0):
         mod = offset
         while True:
             try:
-                next_setting_dt = self.location.sunset(
-                    (self.now + datetime.timedelta(seconds=offset) + datetime.timedelta(days=mod)).date(), local=False
-                )
+                dt = (self.now + datetime.timedelta(seconds=offset) + datetime.timedelta(days=mod)).date()
+                next_setting_dt = self.location.sunset(date=dt, local=False, observer_elevation=self.AD.elevation)
                 if next_setting_dt > self.now:
                     break
-            except astral.AstralError:
+            except ValueError:
                 pass
             mod += 1
 
@@ -235,8 +286,7 @@ class Scheduler:
     @staticmethod
     def get_offset(kwargs):
         if "offset" in kwargs["kwargs"]:
-            if "random_start" in kwargs["kwargs"] \
-                    or "random_end" in kwargs["kwargs"]:
+            if "random_start" in kwargs["kwargs"] or "random_end" in kwargs["kwargs"]:
                 raise ValueError(
                     "Can't specify offset as well as 'random_start' or "
                     "'random_end' in 'run_at_sunrise()' or 'run_at_sunset()'"
@@ -247,12 +297,12 @@ class Scheduler:
             rbefore = kwargs["kwargs"].get("random_start", 0)
             rafter = kwargs["kwargs"].get("random_end", 0)
             offset = random.randint(rbefore, rafter)
-            #self.logger.debug("get_offset(): offset = %s", offset)
+            # self.logger.debug("get_offset(): offset = %s", offset)
         return offset
 
     async def insert_schedule(self, name, aware_dt, callback, repeat, type_, **kwargs):
 
-        #aware_dt will include a timezone of some sort - convert to utc timezone
+        # aware_dt will include a timezone of some sort - convert to utc timezone
         utc = aware_dt.astimezone(pytz.utc)
 
         # Round to nearest second
@@ -289,7 +339,7 @@ class Scheduler:
             "type": type_,
             "pin_app": pin_app,
             "pin_thread": pin_thread,
-            "kwargs": kwargs
+            "kwargs": kwargs,
         }
 
         if callback is None:
@@ -297,21 +347,23 @@ class Scheduler:
         else:
             function_name = callback.__name__
 
-        await self.AD.state.add_entity("admin",
-                                       "scheduler_callback.{}".format(handle),
-                                       "active",
-                                       {
-                                           "app": name,
-                                           "execution_time": utils.dt_to_str(ts.replace(microsecond=0), self.AD.tz),
-                                           "repeat": str(datetime.timedelta(seconds=interval)),
-                                           "function": function_name,
-                                           "pinned": pin_app,
-                                           "pinned_thread": pin_thread,
-                                           "fired": 0,
-                                           "executed": 0,
-                                           "kwargs": kwargs
-                                       })
-                # verbose_log(conf.logger, "INFO", conf.schedule[name][handle])
+        await self.AD.state.add_entity(
+            "admin",
+            "scheduler_callback.{}".format(handle),
+            "active",
+            {
+                "app": name,
+                "execution_time": utils.dt_to_str(ts.replace(microsecond=0), self.AD.tz),
+                "repeat": str(datetime.timedelta(seconds=interval)),
+                "function": function_name,
+                "pinned": pin_app,
+                "pinned_thread": pin_thread,
+                "fired": 0,
+                "executed": 0,
+                "kwargs": kwargs,
+            },
+        )
+        # verbose_log(conf.logger, "INFO", conf.schedule[name][handle])
 
         if self.active is True:
             await self.kick()
@@ -332,22 +384,63 @@ class Scheduler:
 
     def get_next_entries(self):
 
-        next_exec = datetime.datetime.now(pytz.utc).replace(year=3200)
+        next_exec = datetime.datetime.now(pytz.utc).replace(year=datetime.MAXYEAR, month=12, day=31)
         for name in self.schedule.keys():
             for entry in self.schedule[name].keys():
                 if self.schedule[name][entry]["timestamp"] < next_exec:
                     next_exec = self.schedule[name][entry]["timestamp"]
 
-        next_entries =[]
+        next_entries = []
 
         for name in self.schedule.keys():
             for entry in self.schedule[name].keys():
                 if self.schedule[name][entry]["timestamp"] == next_exec:
-                    next_entries.append({"name": name, "uuid": entry, "timestamp": self.schedule[name][entry]["timestamp"]})
+                    next_entries.append(
+                        {"name": name, "uuid": entry, "timestamp": self.schedule[name][entry]["timestamp"]}
+                    )
 
         return next_entries
 
-    async def loop(self):
+    async def process_dst(self, old, new):
+        #
+        # Rewrite timestamps to new local time
+        #
+        offset = old - new
+        self.logger.debug("Process_dst()")
+        self.logger.debug("offset  %s", offset)
+        for app in self.schedule:
+            for entry in self.schedule[app]:
+                args = self.schedule[app][entry]
+                # Sunrise and sunset will already be correct. Anything else needs to be reset to a new local time
+                self.logger.debug("Before rewrite: %s", args)
+                if args["type"] != "next_rising" and args["type"] != "next_setting":
+                    # If our interval is less than the jump don't rewrite the timestamp
+                    if float(args["interval"]) > abs(offset.total_seconds()):
+                        args["timestamp"] += offset
+                        args["basetime"] += offset
+                self.logger.debug("After rewrite: %s", args)
+
+    def get_next_dst_offset(self, base, limit):
+        #
+        # I can't believe there isn't a better way to find the next DST transition but ...
+        # We know the lower and upper bounds of DST so do a search to find the actual transition time
+        # I don't want to rely on heuristics such as "it occurs at 2am" because I don't know if that holds
+        # true for every timezone. With this method, as long as pytz's dst() function is correct, this should work
+        #
+
+        # TODO: Convert this to some sort of binary search for efficiency
+        # TODO: This really should support sub 1 second periods better
+        self.logger.debug("get_next_dst_offset() base=%s limit=%s", base, limit)
+        current = base.astimezone(self.AD.tz).dst()
+        self.logger.debug("current=%s", current)
+        for offset in range(1, int(limit) + 1):
+            candidate = (base + timedelta(seconds=offset)).astimezone(self.AD.tz)
+            # print(candidate)
+            if candidate.dst() != current:
+                return offset
+        return limit
+
+    async def loop(self):  # noqa: C901
         self.active = True
         self.logger.debug("Starting scheduler loop()")
         self.AD.booted = await self.get_now_naive()
@@ -367,7 +460,9 @@ class Scheduler:
 
         next_entries = []
         result = False
-        idle_time = 60
+        idle_time = 1
+        delay = 0
+        old_dst_offset = (await self.get_now()).astimezone(self.AD.tz).dst()
         while not self.stopping:
             try:
                 if self.endtime is not None and self.now >= self.endtime:
@@ -375,9 +470,6 @@ class Scheduler:
                     if self.AD.stop_function is not None:
                         self.AD.stop_function()
                     else:
-                        #
-                        # We aren't in a standalone environment so the best we can do is terminate the AppDaemon parts
-                        #
                         self.stop()
                 now = pytz.utc.localize(datetime.datetime.utcnow())
                 if self.realtime is True:
@@ -389,7 +481,7 @@ class Scheduler:
                     else:
                         if len(next_entries) > 0:
                             # Time is progressing infinitely fast and it's already time for our next callback
-                            delta = (next_entries[0]["timestamp"] - self.now).total_seconds()
+                            delta = delay
                         else:
                             # No kick, no scheduler expiry ...
                             delta = idle_time
@@ -399,6 +491,28 @@ class Scheduler:
                 self.last_fired = pytz.utc.localize(datetime.datetime.utcnow())
                 self.logger.debug("self.now = %s", self.now)
                 #
+                # Now we're awake and know what time it is
+                #
+                dst_offset = (await self.get_now()).astimezone(self.AD.tz).dst()
+                self.logger.debug(
+                    "local now=%s old_dst_offset=%s new_dst_offset=%s",
+                    self.now.astimezone(self.AD.tz),
+                    old_dst_offset,
+                    dst_offset,
+                )
+                if old_dst_offset != dst_offset:
+                    #
+                    # DST began or ended, we need to go fix any existing scheduler entries to match the new local time
+                    #
+                    self.logger.info("Daylight Savings Time transition detected - rewriting events to new local time")
+                    await self.process_dst(old_dst_offset, dst_offset)
+                    #
+                    # Re calculate next entries
+                    #
+                    next_entries = self.get_next_entries()
+
+                old_dst_offset = dst_offset
+                #
                 # OK, lets fire the entries
                 #
                 for entry in next_entries:
@@ -426,20 +540,47 @@ class Scheduler:
                     # Nothing to do, lets wait for a while, we will get woken up if anything new comes along
                     delay = idle_time
 
+                # Initially we don't want to skip over any events that haven;t had a chance to be registered yet, but now
+                # we can loosen up a little
+                idle_time = 60
+
+                #
+                # We are about to go to sleep, but we need to ensure we don't miss a DST transition or we will
+                # sleep in and potentially miss an event that should happen earlier than expected due to the time change
+                #
+
+                next = self.now + timedelta(seconds=delay)
+
+                self.logger.debug("next event=%s", next)
+
+                if await self.is_dst() != await self.is_dst(next):
+                    #
+                    # Reset delay to wake up at the DST change so we can re-jig everything
+                    #
+
+                    delay = self.get_next_dst_offset(self.now, delay)
+                    self.logger.debug(
+                        "DST transition before next event: %s %s", await self.is_dst(), await self.is_dst(next)
+                    )
+
                 self.logger.debug("Delay = %s seconds", delay)
+
                 if delay > 0 and self.AD.timewarp > 0:
+                    #
+                    # Sleep until the next event
+                    #
                     result = await self.sleep(delay / self.AD.timewarp)
                     self.logger.debug("result = %s", result)
                 else:
                     # Not sleeping but lets be fair to the rest of AD
                     await asyncio.sleep(0)
 
-            except:
-                self.logger.warning('-' * 60)
+            except Exception:
+                self.logger.warning("-" * 60)
                 self.logger.warning("Unexpected error in scheduler loop")
-                self.logger.warning('-' * 60)
+                self.logger.warning("-" * 60)
                 self.logger.warning(traceback.format_exc())
-                self.logger.warning('-' * 60)
+                self.logger.warning("-" * 60)
                 # Prevent spamming of the logs
                 await self.sleep(1)
 
@@ -455,7 +596,7 @@ class Scheduler:
 
     async def kick(self):
         while self.sleep_task is None:
-            await asyncio.sleep(0)
+            await asyncio.sleep(1)
         self.sleep_task.cancel()
 
     #
@@ -469,12 +610,12 @@ class Scheduler:
         return self.next_sunrise() < self.next_sunset()
 
     async def info_timer(self, handle, name):
-        if name in self.schedule and handle in self.schedule[name]:
+        if self.timer_running(name, handle):
             callback = self.schedule[name][handle]
             return (
                 self.make_naive(callback["timestamp"]),
                 callback["interval"],
-                self.sanitize_timer_kwargs(self.AD.app_management.objects[name]["object"], callback["kwargs"])
+                self.sanitize_timer_kwargs(self.AD.app_management.objects[name]["object"], callback["kwargs"]),
             )
         else:
             self.logger.warning("Invalid timer handle given as: %s", handle)
@@ -484,22 +625,29 @@ class Scheduler:
         schedule = {}
         for name in self.schedule.keys():
             schedule[name] = {}
-            for entry in sorted(
-                    self.schedule[name].keys(),
-                    key=lambda uuid_: self.schedule[name][uuid_]["timestamp"]
-            ):
+            for entry in sorted(self.schedule[name].keys(), key=lambda uuid_: self.schedule[name][uuid_]["timestamp"],):
                 schedule[name][str(entry)] = {}
-                schedule[name][str(entry)]["timestamp"] = str(self.AD.sched.make_naive(self.schedule[name][entry]["timestamp"]))
+                schedule[name][str(entry)]["timestamp"] = str(
+                    self.AD.sched.make_naive(self.schedule[name][entry]["timestamp"])
+                )
                 schedule[name][str(entry)]["type"] = self.schedule[name][entry]["type"]
                 schedule[name][str(entry)]["name"] = self.schedule[name][entry]["name"]
-                schedule[name][str(entry)]["basetime"] = str(self.AD.sched.make_naive(self.schedule[name][entry]["basetime"]))
+                schedule[name][str(entry)]["basetime"] = str(
+                    self.AD.sched.make_naive(self.schedule[name][entry]["basetime"])
+                )
                 schedule[name][str(entry)]["repeat"] = self.schedule[name][entry]["repeat"]
                 if self.schedule[name][entry]["type"] == "next_rising":
-                    schedule[name][str(entry)]["interval"] = "sunrise:{}".format(utils.format_seconds(self.schedule[name][entry]["offset"]))
+                    schedule[name][str(entry)]["interval"] = "sunrise:{}".format(
+                        utils.format_seconds(self.schedule[name][entry]["offset"])
+                    )
                 elif self.schedule[name][entry]["type"] == "next_setting":
-                    schedule[name][str(entry)]["interval"] = "sunset:{}".format(utils.format_seconds(self.schedule[name][entry]["offset"]))
+                    schedule[name][str(entry)]["interval"] = "sunset:{}".format(
+                        utils.format_seconds(self.schedule[name][entry]["offset"])
+                    )
                 elif self.schedule[name][entry]["repeat"] is True:
-                    schedule[name][str(entry)]["interval"] = utils.format_seconds(self.schedule[name][entry]["interval"])
+                    schedule[name][str(entry)]["interval"] = utils.format_seconds(
+                        self.schedule[name][entry]["interval"]
+                    )
                 else:
                     schedule[name][str(entry)]["interval"] = "None"
 
@@ -508,8 +656,14 @@ class Scheduler:
                 for kwarg in self.schedule[name][entry]["kwargs"]:
                     schedule[name][str(entry)]["kwargs"] = utils.get_kwargs(self.schedule[name][entry]["kwargs"])
                 schedule[name][str(entry)]["callback"] = self.schedule[name][entry]["callback"].__name__
-                schedule[name][str(entry)]["pin_thread"] = self.schedule[name][entry]["pin_thread"] if self.schedule[name][entry]["pin_thread"] != -1 else "None"
-                schedule[name][str(entry)]["pin_app"] = "True" if self.schedule[name][entry]["pin_app"] is True else "False"
+                schedule[name][str(entry)]["pin_thread"] = (
+                    self.schedule[name][entry]["pin_thread"]
+                    if self.schedule[name][entry]["pin_thread"] != -1
+                    else "None"
+                )
+                schedule[name][str(entry)]["pin_app"] = (
+                    "True" if self.schedule[name][entry]["pin_app"] is True else "False"
+                )
 
         # Order it
 
@@ -517,8 +671,11 @@ class Scheduler:
 
         return ordered_schedule
 
-    async def is_dst(self):
-        return (await self.get_now()).astimezone(self.AD.tz).dst() != datetime.timedelta(0)
+    async def is_dst(self, dt=None):
+        if dt is None:
+            return (await self.get_now()).astimezone(self.AD.tz).dst() != datetime.timedelta(0)
+        else:
+            return dt.astimezone(self.AD.tz).dst() != datetime.timedelta(0)
 
     async def get_now(self):
         if self.realtime is True:
@@ -543,13 +700,8 @@ class Scheduler:
         start_time = (await self._parse_time(start_time_str, name))["datetime"]
         end_time = (await self._parse_time(end_time_str, name))["datetime"]
         now = (await self.get_now()).astimezone(self.AD.tz)
-        start_date = now.replace(
-            hour=start_time.hour, minute=start_time.minute,
-            second=start_time.second
-        )
-        end_date = now.replace(
-            hour=end_time.hour, minute=end_time.minute, second=end_time.second
-        )
+        start_date = now.replace(hour=start_time.hour, minute=start_time.minute, second=start_time.second)
+        end_date = now.replace(hour=end_time.hour, minute=end_time.minute, second=end_time.second)
         if end_date < start_date:
             # Spans midnight
             if now < start_date and now < end_date:
@@ -581,23 +733,28 @@ class Scheduler:
         else:
             return self.make_naive((await self._parse_time(time_str, name))["datetime"])
 
-
     async def _parse_time(self, time_str, name=None):
         parsed_time = None
         sun = None
         offset = 0
-        parts = re.search('^(\d+)-(\d+)-(\d+)\s+(\d+):(\d+):(\d+)$', time_str)
+        parts = re.search(r"^(\d+)-(\d+)-(\d+)\s+(\d+):(\d+):(\d+)$", time_str)
         if parts:
-            this_time = datetime.datetime(int(parts.group(1)), int(parts.group(2)), int(parts.group(3)), int(parts.group(4)), int(parts.group(5)), int(parts.group(6)), 0)
+            this_time = datetime.datetime(
+                int(parts.group(1)),
+                int(parts.group(2)),
+                int(parts.group(3)),
+                int(parts.group(4)),
+                int(parts.group(5)),
+                int(parts.group(6)),
+                0,
+            )
             parsed_time = self.AD.tz.localize(this_time)
         else:
-            parts = re.search('^(\d+):(\d+):(\d+)$', time_str)
+            parts = re.search(r"^(\d+):(\d+):(\d+)$", time_str)
             if parts:
                 today = (await self.get_now()).astimezone(self.AD.tz)
-                time = datetime.time(
-                    int(parts.group(1)), int(parts.group(2)), int(parts.group(3)), 0
-                )
-                parsed_time = today.replace(hour=time.hour, minute=time.minute, second=time.second, microsecond=0)
+                time = datetime.time(int(parts.group(1)), int(parts.group(2)), int(parts.group(3)), 0)
+                parsed_time = today.replace(hour=time.hour, minute=time.minute, second=time.second, microsecond=0,)
 
             else:
                 if time_str == "sunrise":
@@ -609,49 +766,40 @@ class Scheduler:
                     sun = "sunset"
                     offset = 0
                 else:
-                    parts = re.search(
-                        '^sunrise\s*([+-])\s*(\d+):(\d+):(\d+)$', time_str
-                    )
+                    parts = re.search(r"^sunrise\s*([+-])\s*(\d+):(\d+):(\d+)$", time_str)
                     if parts:
                         sun = "sunrise"
                         if parts.group(1) == "+":
                             td = datetime.timedelta(
-                                hours=int(parts.group(2)), minutes=int(parts.group(3)),
-                                seconds=int(parts.group(4))
+                                hours=int(parts.group(2)), minutes=int(parts.group(3)), seconds=int(parts.group(4)),
                             )
                             offset = td.total_seconds()
-                            parsed_time = (await self.sunrise(True) + td)
+                            parsed_time = await self.sunrise(True) + td
                         else:
                             td = datetime.timedelta(
-                                hours=int(parts.group(2)), minutes=int(parts.group(3)),
-                                seconds=int(parts.group(4))
+                                hours=int(parts.group(2)), minutes=int(parts.group(3)), seconds=int(parts.group(4)),
                             )
                             offset = td.total_seconds() * -1
-                            parsed_time = (await self.sunrise(True) - td)
+                            parsed_time = await self.sunrise(True) - td
                     else:
-                        parts = re.search(
-                            '^sunset\s*([+-])\s*(\d+):(\d+):(\d+)$', time_str
-                        )
+                        parts = re.search(r"^sunset\s*([+-])\s*(\d+):(\d+):(\d+)$", time_str)
                         if parts:
                             sun = "sunset"
                             if parts.group(1) == "+":
                                 td = datetime.timedelta(
-                                    hours=int(parts.group(2)), minutes=int(parts.group(3)),
-                                    seconds=int(parts.group(4))
+                                    hours=int(parts.group(2)), minutes=int(parts.group(3)), seconds=int(parts.group(4)),
                                 )
                                 offset = td.total_seconds()
-                                parsed_time = (await self.sunset(True) + td)
+                                parsed_time = await self.sunset(True) + td
                             else:
                                 td = datetime.timedelta(
-                                    hours=int(parts.group(2)), minutes=int(parts.group(3)),
-                                    seconds=int(parts.group(4))
+                                    hours=int(parts.group(2)), minutes=int(parts.group(3)), seconds=int(parts.group(4)),
                                 )
                                 offset = td.total_seconds() * -1
-                                parsed_time = (await self.sunset(True) - td)
+                                parsed_time = await self.sunset(True) - td
         if parsed_time is None:
             if name is not None:
-                raise ValueError(
-                    "%s: invalid time string: %s", name, time_str)
+                raise ValueError("%s: invalid time string: %s", name, time_str)
             else:
                 raise ValueError("invalid time string: %s", time_str)
         return {"datetime": parsed_time, "sun": sun, "offset": offset}
@@ -678,10 +826,13 @@ class Scheduler:
             for name in self.schedule.keys():
                 self.diag.info("%s:", name)
                 for entry in sorted(
-                        self.schedule[name].keys(),
-                        key=lambda uuid_: self.schedule[name][uuid_]["timestamp"]
+                    self.schedule[name].keys(), key=lambda uuid_: self.schedule[name][uuid_]["timestamp"],
                 ):
-                    self.diag.info(" Next Event Time: %s - data: %s", self.make_naive(self.schedule[name][entry]["timestamp"]), self.schedule[name][entry])
+                    self.diag.info(
+                        " Next Event Time: %s - data: %s",
+                        self.make_naive(self.schedule[name][entry]["timestamp"]),
+                        self.schedule[name][entry],
+                    )
             self.diag.info("--------------------------------------------------")
 
     #
@@ -690,9 +841,11 @@ class Scheduler:
     @staticmethod
     def sanitize_timer_kwargs(app, kwargs):
         kwargs_copy = kwargs.copy()
-        return utils._sanitize_kwargs(kwargs_copy, [
-            "interval", "constrain_days", "constrain_input_boolean", "_pin_app", "_pin_thread"
-        ] + app.list_constraints())
+        return utils._sanitize_kwargs(
+            kwargs_copy,
+            ["interval", "constrain_days", "constrain_input_boolean", "_pin_app", "_pin_thread", "__silent"]
+            + app.list_constraints(),
+        )
 
     @staticmethod
     def myround(x, base=1, prec=10):
@@ -716,7 +869,7 @@ class Scheduler:
         # Is it naive?
         result = None
         if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
-            #Localize with the configured timezone
+            # Localize with the configured timezone
             result = self.AD.tz.localize(dt)
         else:
             result = dt
@@ -725,4 +878,6 @@ class Scheduler:
 
     def make_naive(self, dt):
         local = dt.astimezone(self.AD.tz)
-        return datetime.datetime(local.year, local.month, local.day,local.hour, local.minute, local.second, local.microsecond)
+        return datetime.datetime(
+            local.year, local.month, local.day, local.hour, local.minute, local.second, local.microsecond,
+        )
